use std::error::Error;

use super::{Number, Sign};

impl TryFrom<Number> for isize {
    type Error = Box<dyn Error>;

    fn try_from(value: Number) -> Result<Self, Self::Error> {
        let mut num = 0;
        for (pos, &digit) in value.digits.iter().rev().enumerate() {
            let mul = digit.checked_mul(10isize.pow(pos as u32));
            if mul.is_none() {
                return Err(Box::from("Cannot convert Number to isize. Too big."));
            }
            num += mul.unwrap();
        }
        Ok(num)
    }
}

impl From<Number> for String {
    fn from(value: Number) -> Self {
        let string_vec: Vec<String> = value
            .digits
            .iter()
            .map(|&digit| digit.to_string())
            .collect();
        string_vec.concat()
    }
}

impl From<&str> for Number {
    fn from(value: &str) -> Self {
        let bytes = value.as_bytes();
        let (sign, idx_start) = match bytes[0] {
            b'-' => (Sign::Negatif, 1),
            _ => (Sign::Positif, 0),
        };
        let mut digits = vec![];
        for &byte in &bytes[idx_start..] {
            let digit = Self::byte_to_digit(byte);
            digits.push(digit);
        }
        Self { digits, sign }
    }
}

impl From<isize> for Number {
    fn from(mut value: isize) -> Self {
        let mut sign = Sign::Positif;
        if value < 0 {
            sign = Sign::Negatif;
            value *= -1;
        }
        if value == 0 {
            return Self {
                digits: vec![0],
                sign: Sign::Positif,
            };
        }
        let num_len = (value as f64 + 1.0).log10().ceil() as usize;
        let mut digits = vec![];
        for digit_idx in 0..num_len {
            let digit = Self::get_digit(value, digit_idx);
            digits.push(digit);
        }
        let digits = digits.iter().rev().copied().collect();
        Self { digits, sign }
    }
}

#[cfg(test)]
mod test_number_from {
    use crate::number::{Number, Sign};

    #[test]
    fn test_from_isize() {
        let number = Number::from(-1234);
        assert_eq!(
            number,
            Number {
                digits: vec![1, 2, 3, 4],
                sign: Sign::Negatif
            }
        );
    }

    #[test]
    fn test_from_0isize() {
        let number = Number::from(0);
        assert_eq!(
            number,
            Number {
                digits: vec![0],
                sign: Sign::Positif
            }
        );
    }
}
